using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;

namespace Syntriax.Engine.Core.Abstract;

/// <summary>
/// Represents a controller for managing <see cref="IBehaviour"/>s and notify them accordingly about the engine's updates. Connected to an <see cref="IGameObject"/>.
/// </summary>
public interface IBehaviourController : IInitialize, IAssignableGameObject, IEnumerable<IBehaviour>
{
    /// <summary>
    /// Event triggered before the update of <see cref="IBehaviour"/>s.
    /// </summary>
    event OnPreUpdateEventHandler? OnPreUpdate;

    /// <summary>
    /// Event triggered during the update of <see cref="IBehaviour"/>s.
    /// </summary>
    event OnUpdateEventHandler? OnUpdate;

    /// <summary>
    /// Event triggered before the drawing phase.
    /// </summary>
    event OnPreDrawEventHandler? OnPreDraw;

    /// <summary>
    /// Event triggered when a <see cref="IBehaviour"/> is added to the <see cref="IBehaviourController"/>.
    /// </summary>
    event OnBehaviourAddedEventHandler? OnBehaviourAdded;

    /// <summary>
    /// Event triggered when a <see cref="IBehaviour"/> is removed from the <see cref="IBehaviourController"/>.
    /// </summary>
    event OnBehaviourRemovedEventHandler? OnBehaviourRemoved;

    /// <summary>
    /// Adds a <see cref="IBehaviour"/> to the <see cref="IBehaviourController"/>.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/> to add.</typeparam>
    /// <param name="behaviour">The <see cref="IBehaviour"/> to add.</param>
    /// <returns>The added <see cref="IBehaviour"/>.</returns>
    T AddBehaviour<T>(T behaviour) where T : class, IBehaviour;

    /// <summary>
    /// Adds a <see cref="IBehaviour"/> of the specified type to the <see cref="IBehaviourController"/>.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/> to add.</typeparam>
    /// <param name="args">Construction parameters for the <see cref="IBehaviour"/>.</param>
    /// <returns>The added <see cref="IBehaviour"/>.</returns>
    T AddBehaviour<T>(params object?[]? args) where T : class, IBehaviour;

    /// <summary>
    /// Gets a <see cref="IBehaviour"/> of the specified type.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/> to get.</typeparam>
    /// <returns>The <see cref="IBehaviour"/> of the specified type if found; otherwise, <see cref="null"/>.</returns>
    T? GetBehaviour<T>();

    /// <summary>
    /// Gets all <see cref="IBehaviour"/>s of the specified type.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/>s to get.</typeparam>
    /// <returns>A list of <see cref="IBehaviour"/>s of the specified type.</returns>
    IList<T> GetBehaviours<T>();

    /// <summary>
    /// Gets all <see cref="IBehaviour"/>s of the specified type and stores them in the provided list.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/>s to get.</typeparam>
    /// <param name="results">The list to store the <see cref="IBehaviour"/>s.</param>
    void GetBehaviours<T>(IList<T> results);

    /// <summary>
    /// Removes <see cref="IBehaviour"/>s of the specified type from the <see cref="IBehaviourController"/>.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/>s to remove.</typeparam>
    /// <param name="removeAll">A flag indicating whether to remove all <see cref="IBehaviour"/>s of the specified type.</param>
    void RemoveBehaviour<T>(bool removeAll = false) where T : class, IBehaviour;

    /// <summary>
    /// Removes the specified <see cref="IBehaviour"/> from the <see cref="IBehaviourController"/>.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/> to remove.</typeparam>
    /// <param name="behaviour">The <see cref="IBehaviour"/> to remove.</param>
    void RemoveBehaviour<T>(T behaviour) where T : class, IBehaviour;

    /// <summary>
    /// Updates all <see cref="IBehaviour"/>s in the <see cref="IBehaviourController"/>.
    /// </summary>
    void Update();

    /// <summary>
    /// Performs pre-draw operations.
    /// </summary>
    void UpdatePreDraw();

    delegate void OnPreUpdateEventHandler(IBehaviourController sender);
    delegate void OnUpdateEventHandler(IBehaviourController sender);
    delegate void OnPreDrawEventHandler(IBehaviourController sender);
    delegate void OnBehaviourAddedEventHandler(IBehaviourController sender, IBehaviour behaviourAdded);
    delegate void OnBehaviourRemovedEventHandler(IBehaviourController sender, IBehaviour behaviourRemoved);

}
